
from typing import List
from itertools import product

import numpy as np
import numpy.typing as npt
import pandas as pd
from MDAnalysis.analysis.base import AnalysisBase
from MDAnalysis.lib.distances import distance_array, calc_angles, calc_bonds

from ectoolkits.log import get_logger

logger = get_logger(__name__)

class HydrogenBondAnalysis(AnalysisBase):
    """

    """
    def __init__(self,
                 atomgroup,
                 idxs_target: List[int],
                 idxs_other: List[int],
                 type_HB="accept",
                 verbose=True,
                 **kwargs):
        """
        Initialize the ProtonTransferCV analysis.

        Parameters:
            atomgroup (AtomGroup): The atomgroup to analyze.
            idxs_target (List[int]): List of indices for the target atoms (one per frame).
            idxs_other (List[int]): List of indices for the other atoms that potentially donate or accept HBs
            (idxs_other is applied to the whole trajectory).
            verbose (bool): Whether to print verbose output. Defaults to True.
            **kwargs: Additional keyword arguments.


        Set up the initial analysis parameters.
        I don't know the reason why the offical example always put atomgroup as the first argument.
        Universe class can also access the atomgroup and trajectory classes.
        """
        trajectory = atomgroup.universe.trajectory
        super(HydrogenBondAnalysis, self).__init__(trajectory, verbose=verbose)

        logger.info("Analysis of Hydrogen Bonds")
        logger.info(f"Analysing the number of {type_HB} hydrogen bonds ")

        self.u = atomgroup.universe

        self.idxs_target = idxs_target
        self.idxs_other = idxs_other
        self.type_HB = type_HB
 

        # How should I select atoms inside AnalysisBase?
        # Now I just use the Universe object to select atoms.
        self.ag_O = self.u.select_atoms('type O')
        self.ag_H = self.u.select_atoms('type H')

        self.n_results = 2



    def _prepare(self):
        if self.n_frames != len(self.idxs_target):
            raise ValueError("The number of frames and the number of target atoms should be the same.")
        
        self.results = np.zeros((self.n_frames, self.n_results))

    def _single_frame(self):

        idx_target = self.idxs_target[self._frame_index]

        if np.isnan(idx_target):
            self.results[self._frame_index] = np.array([self._ts.frame, np.nan])
        else:
            idxs_OH_pair = self._gen_connectivity()

            if self.type_HB == "accept":

                # type1 as donor and type2 as acceptor
                #   candidates_hbonds = [ (donor_idx, hydrogen_idx, acceptor_idx..)]
                candidates_hbonds = self._get_candidates_hbonds(self.idxs_other, [idx_target], idxs_OH_pair)
                

            elif self.type_HB == "donate":
                # candidates_hbonds = [ (donor_idx, hydrogen_idx, acceptor_idx..)]
                candidates_hbonds = self._get_candidates_hbonds([idx_target], self.idxs_other, idxs_OH_pair)

            info = self._get_num_hbonds(candidates_hbonds)


            self.results[self._frame_index] = info



    def _conclude(self):
        """
        Finish up by calculating an average and transforming our
        results into a DataFrame.
        """
        # by now self.result is fully populated
        # [frame, delta_cv, distance_cv, donor_index, hydrogen_index, acceptor_index, dOdHd, dOaHd, DA_distance, DAH_angle ]
        columns = ['Frame', f'num_{self.type_HB}_hbonds']

        self.df = pd.DataFrame(self.results, columns=columns)

    # helper functions
    def _gen_connectivity(self):
        """
        Generate the connectivity between O and H atoms.
        """
        poses_O = self.ag_O.positions
        idxs_O = self.ag_O.indices
        poses_H = self.ag_H.positions
        idxs_H = self.ag_H.indices
        # We define each H atom to be “covalently bound” to its nearest O atom.
        # See 10.1021/acs.jpclett.7b00358
        dist_array = distance_array(poses_H, poses_O, box=self.u.dimensions)
        idxs_O_pair = idxs_O[np.argmin(dist_array, axis=1)]
        idxs_H_pair = idxs_H
        idxs_OH_pair = np.stack((idxs_O_pair, idxs_H_pair), axis=1)
        return idxs_OH_pair

    def _get_candidates_hbonds(self,
                               idxs_donor: List[int],
                               idxs_acceptor: List[int],
                               idxs_OH_pair: npt.NDArray,
                               ):
        """
        Get the candidates of hydrogen bonds.
        """

        idxs_O_pair = idxs_OH_pair[:, 0]
        idxs_H_pair = idxs_OH_pair[:, 1]

        # fisrt make product of donor and acceptor
        # insert the permuation of bridge water oxygen.
        # and make the candidates_hbonds_list in the order of donor0, hydrogen0, donor1, hydrogen1, ..., acceptor
        candidates_hbonds = []


        for idx_donor, idx_acceptor in product(idxs_donor, idxs_acceptor):
            idxs_bonded_H = idxs_H_pair[idxs_O_pair == idx_donor]
            for idx_bonded_H in idxs_bonded_H:
                candidates_hbonds.append([idx_donor, idx_bonded_H, idx_acceptor])

        return np.array(candidates_hbonds)

    def _get_num_hbonds(self, candidates_hbonds):

        # hbond standard
        hbond_dOO = 3.5
        hbond_OaOdH_angle = 30


        info = np.zeros((self.n_results))
        info[0] = self._ts.frame

        if candidates_hbonds.shape[0] == 0:
            info[1:] = np.nan
        else:
            # check if the h bond forms
            # check hbond distance first
            # donor_index_0, hydrogen_index_0, donor_index_1, hydrogen_index_1,
            # ..., donor_index_num_bridge, hydrogen_index_num_bridge, acceptor_index
            num_candidates = candidates_hbonds.shape[0]
            mask = np.full(num_candidates, True)

            poses_donor_o = self.u.atoms.positions[candidates_hbonds.T[0]]
            poses_acceptor_o = self.u.atoms.positions[candidates_hbonds.T[2]]
            bonds = calc_bonds(poses_acceptor_o, poses_donor_o, box=self.u.dimensions)
            mask = mask & (bonds < hbond_dOO)

            candidates_hbonds = candidates_hbonds[mask]


            # then check the angle
            num_candidates = candidates_hbonds.shape[0]
            mask = np.full(num_candidates, True)

            poses_donor_o = self.u.atoms.positions[candidates_hbonds.T[0]]
            poses_bonded_H = self.u.atoms.positions[candidates_hbonds.T[1]]
            poses_acceptor_o = self.u.atoms.positions[candidates_hbonds.T[2]]
            angles = calc_angles(poses_acceptor_o, poses_donor_o, poses_bonded_H, box=self.u.dimensions)/np.pi*180
            mask = mask & (angles <= hbond_OaOdH_angle)

            # All the candidates have hydrogen bond formed.
            hbonds = candidates_hbonds[mask]
            info[1]= len(hbonds)

        return info



from ase.io import read
from MDAnalysis import Universe
from MDAnalysis.coordinates.chain import ChainReader


for i in range(4):
    dump_files = [f"/capstor/store/cscs/userlab/lp07/zyongbin/BiVO4FEP/05.metad/opes_new/opes_pot{i:d}/bak_traj/{(j*5):02d}-{((j+1)*5):02d}ns.dump.lammpstrj" for j in range(6)]
    print("Processing", flush=True)
    for dump_file in dump_files:
          print(f"    {dump_file}", flush=True)

    stc = read(dump_files[0], format="lammps-dump-text", specorder=["Bi", "H", "O", "V"])
    chemical_symbols = stc.get_chemical_symbols()


    u= Universe(dump_files[0], format="LAMMPSDUMP")
    u.trajectory = ChainReader(dump_files, format='LAMMPSDUMP')

    u.add_TopologyAttr("types", np.array(chemical_symbols))

    print(len(u.trajectory))

    atomgroup = u.atoms

    idxs_water = [144, 151, 157, 162, 168, 169, 180, 183, 185, 186, 187, 188, 189, 191, 216, 217, 222, 223, 224, 226, 227, 228, 229, 232, 233, 234, 235, 236, 239, 240, 242, 243, 244, 247, 315, 321, 324, 327, 330, 339, 342, 345, 348, 351, 354, 130, 145, 156, 163, 174, 175, 181, 182, 184, 190, 218, 219, 220, 221, 225, 230, 231, 237, 238, 245, 246, 312, 318, 333, 336, 357]

    ptcv_data = np.load(f"/capstor/store/cscs/userlab/lp07/zyongbin/BiVO4FEP/18.PT_CV/IndirectPTOsi-mlp{i:d}.npy")

    # the number of donated HB to PT donor 
    idxs_target = ptcv_data[:, 3]
    idxs_target = [int(x) if not np.isnan(x) else x for x in idxs_target]

    hb = HydrogenBondAnalysis(atomgroup=atomgroup, idxs_target=idxs_target, idxs_other=idxs_water, type_HB="donate", verbose=True)
    hb.run()

    np.save(f"./n_donate_hbond_to_donor_IndirectPTOsi_mlp{i:d}.npy", hb.results)

    # the number of acceptd HB to PT donor 
    idxs_target = ptcv_data[:, 3]
    idxs_target = [int(x) if not np.isnan(x) else x for x in idxs_target]

    hb = HydrogenBondAnalysis(atomgroup=atomgroup, idxs_target=idxs_target, idxs_other=idxs_water, type_HB="accept", verbose=True)
    hb.run()

    np.save(f"./n_accept_hbond_to_donor_IndirectPTOsi_mlp{i:d}.npy", hb.results)

    # the number of donated HB to PT acceptor
    idxs_target = ptcv_data[:, 7]
    idxs_target = [int(x) if not np.isnan(x) else x for x in idxs_target]

    hb = HydrogenBondAnalysis(atomgroup=atomgroup, idxs_target=idxs_target, idxs_other=idxs_water, type_HB="donate", verbose=True)
    hb.run()

    np.save(f"./n_donate_hbond_to_acceptor_IndirectPTOsi_mlp{i:d}.npy", hb.results)


    # the number of accepted HB to PT acceptor
    idxs_target = ptcv_data[:, 7]
    idxs_target = [int(x) if not np.isnan(x) else x for x in idxs_target]

    hb = HydrogenBondAnalysis(atomgroup=atomgroup, idxs_target=idxs_target, idxs_other=idxs_water, type_HB="accept", verbose=True)
    hb.run()

    np.save(f"./n_accept_hbond_to_acceptor_IndirectPTOsi_mlp{i:d}.npy", hb.results)

